from onegov.ticket.handler import Handler, HandlerRegistry
handlers = HandlerRegistry()  # noqa

from onegov.ticket.model import Ticket
from onegov.ticket.collection import TicketCollection


__all__ = [
    'Handler',
    'handlers',
    'Ticket',
    'TicketCollection'
]



from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import JSON, UUID
from onegov.ticket import handlers
from onegov.user.model import User
from sqlalchemy import Column, Enum, ForeignKey, Text
from sqlalchemy.orm import deferred, relationship
from uuid import uuid4


class Ticket(Base, TimestampMixin):
    """ Defines a ticket. """

    __tablename__ = 'tickets'

    #: the internal number of the ticket -> may be used as an access key
    #: for anonymous users
    id = Column(UUID, primary_key=True, default=uuid4)

    #: the unique ticket number known to the end-user -> do *not* use this to
    #: access the ticket as an anonymous user, the number is unique, but it's
    #: not unguessable!
    number = Column(Text, unique=True, nullable=False)

    #: the title of the ticket
    title = Column(Text, nullable=False)

    #: the group this ticket belongs to. used to differentiate tickets
    #: belonging to one specific handler (handler -> group -> title)
    group = Column(Text, nullable=False)

    #: the name of the handler associated with this ticket, may be used to
    #: create custom polymorphic subclasses of this class. See
    #: `<http://docs.sqlalchemy.org/en/improve_toc/\
    #: orm/extensions/declarative/inheritance.html>`_.
    handler_code = Column(Text, nullable=False, index=True)

    #: a unique id for the handler record
    handler_id = Column(Text, nullable=False, index=True, unique=True)

    #: the data associated with the handler, not meant to be loaded in a list,
    #: therefore deferred.
    handler_data = deferred(Column(JSON, nullable=False, default=dict))

    #: a snapshot of the ticket containing the last summary that made any sense
    #: use this before deleting the model behind a ticket, lest your ticket
    #: becomes nothing more than a number.
    snapshot = deferred(Column(JSON, nullable=False, default=dict))

    #: the state of this ticket (open, pending, closed)
    state = Column(
        Enum(
            'open',
            'pending',
            'closed',
            name='ticket_state'
        ),
        nullable=False,
        default='open'
    )

    #: the user that owns this ticket with this ticket (optional)
    user_id = Column(UUID, ForeignKey(User.id), nullable=True)
    user = relationship(User, backref="tickets")

    __mapper_args__ = {
        'polymorphic_on': handler_code
    }

    @property
    def handler(self):
        """ Returns an instance of the handler associated with this ticket. """

        return handlers.get(self.handler_code)(
            self, self.handler_id, self.handler_data)

    def accept_ticket(self, user):
        assert self.state == 'open'

        self.user = user
        self.state = 'pending'

    def close_ticket(self):
        assert self.state == 'pending'

        self.state = 'closed'

    def reopen_ticket(self, user):
        assert self.state == 'closed'

        self.user = user
        self.state = 'pending'

    def create_snapshot(self, request):
        """ Takes the current handler and stores the output of the summary
        as a snapshot.

        TODO: This doesn't support multiple langauges at this point. The
        language of the user creating the snapshot will be what's stored.

        In the future we might change this by iterating over all supported
        languages and creating the summary for each language.

        """

        self.snapshot['summary'] = self.handler.get_summary(request)
        self.snapshot['email'] = self.handler.email



import pytest

from onegov.ticket import Ticket
from onegov.user.model import User


def test_transitions(session):

    ticket = Ticket(state='open')
    assert ticket.state == 'open'
    assert ticket.user is None

    user = User()

    with pytest.raises(AssertionError):
        ticket.close_ticket()
    with pytest.raises(AssertionError):
        ticket.reopen_ticket(user)

    ticket.accept_ticket(user)
    assert ticket.state == 'pending'
    assert ticket.user == user

    with pytest.raises(AssertionError):
        ticket.accept_ticket(user)
    with pytest.raises(AssertionError):
        ticket.reopen_ticket(user)

    ticket.close_ticket()
    assert ticket.state == 'closed'
    assert ticket.user == user

    with pytest.raises(AssertionError):
        ticket.accept_ticket(user)
    with pytest.raises(AssertionError):
        ticket.close_ticket()

    ticket.reopen_ticket(User())
    assert ticket.state == 'pending'
    assert ticket.user != user
    assert ticket.user is not None



from onegov.ticket.errors import DuplicateHandlerError
from sqlalchemy.orm import object_session


class Handler(object):
    """ Defines a generic handler, responsible for a subset of the tickets.

    onegov.ticket is meant to be a rather generic bucket of tickets, to which
    varying modules submit tickets with varying content and different
    actionables.

    Each module wanting to submit tickets needs to implement a handler with
    a unique id and a unique short code. The short code is added in front of
    the ticket number.

    Tickets submitted to the database are shown in a list, without handler
    involvement. When a ticket is displayed, the handler is called with
    whatever data the handler supplied during ticket submission.

    The handler then uses the handler data to access whatever data it needs
    to display a summary as well as links to certain actions (possibly a link
    to the original item, links that change the state of the ticket as well
    as the data associated with the handler, etc.).

    """

    def __init__(self, ticket, handler_id, handler_data):
        self.ticket = ticket
        self.id = handler_id
        self.data = handler_data

    @property
    def session(self):
        return object_session(self.ticket)

    def refresh(self):
        """ Updates the current ticket with the latest data from the handler.
        """

        if self.deleted:
            return

        if self.ticket.title != self.title:
            self.ticket.title = self.title

        if self.ticket.group != self.group:
            self.ticket.group = self.group

        if self.ticket.handler_id != self.id:
            self.ticket.handler_id = self.id

        if self.ticket.handler_data != self.data:
            self.ticket.handler_data = self.data

    @property
    def email(self):
        """ Returns the email address behind the ticket request. """
        raise NotImplementedError

    @property
    def title(self):
        """ Returns the title of the ticket. If this title may change over
        time, the handler must call :meth:`self.refresh` when there's a change.

        """
        raise NotImplementedError

    @property
    def group(self):
        """ Returns the group of the ticket. If this group may change over
        time, the handler must call :meth:`self.refresh` when there's a change.

        """
        raise NotImplementedError

    @property
    def deleted(self):
        """ Returns true if the underlying model was deleted. It is best to
        never let that happen, as we want tickets to stay around forever.

        However, this can make sense in certain scenarios. Note that if
        you do delete your underlying model, make sure to call
        :meth:`onegov.ticket.models.Ticket.create_snapshot` beforehand!

        """

        raise NotImplementedError

    def get_summary(self, request):
        """ Returns the summary of the current ticket as a html string. """

        raise NotImplementedError

    def get_links(self, request):
        """ Returns the links associated with the current ticket in the
        following format::

            [
                ('Link Title', 'http://link'),
                ('Link Title 2', 'http://link2'),
            ]

        If the links are not tuples, but callables, they will be called with
        the request which should return the rendered link.
        """

        raise NotImplementedError


class HandlerRegistry(object):

    _reserved_handler_codes = {'ALL'}

    def __init__(self):
        self.registry = {}

    def get(self, handler_code):
        """ Returns the handler registration for the given id. If the id
        does not exist, a KeyError is raised.

        """
        return self.registry[handler_code]

    def register(self, handler_code, handler_class):
        """ Registers a handler.

        :handler_code:
            Three characters long shortcode of the handler added in front of
            the ticket number. Must be globally unique and must not change!

            Examples for good handler_codes::

                FRM
                ONS
                RES
                EVT

        :handler_class:
            A handler class inheriting from :class:`Handler`.

        """

        assert len(handler_code) == 3
        assert handler_code == handler_code.upper()
        assert handler_code not in self._reserved_handler_codes

        if handler_code in self.registry:
            raise DuplicateHandlerError

        self.registry[handler_code] = handler_class

    def registered_handler(self, handler_code):
        """ A decorator to register handles as follows::

        @handlers.registered_handler('FOO')
        class FooHandler(Handler):
            pass

        """
        def wrapper(handler_class):
            self.register(handler_code, handler_class)
            return handler_class
        return wrapper
